@@ -37,17 +37,23 @@
 #include <getopt.h>
 #include <string.h>
 #include <errno.h>
+#include <io.h>
 #include "imagehelper.h"
+#include "pechecksum.h"
 #include "rebase-db.h"
+#include <versionhelpers.h> /* requires <windows.h> */
 
 BOOL save_image_info ();
 BOOL load_image_info ();
 BOOL merge_image_info ();
 BOOL collect_image_info (const char *pathname);
 void print_image_info ();
+static void print_overlapped ();
 BOOL rebase (const char *pathname, ULONG64 *new_image_base, BOOL down_flag);
 void parse_args (int argc, char *argv[]);
 unsigned long long string_to_ulonglong (const char *string);
+static int compactos_get_algorithm (const char *pathname);
+static int compactos_compress_file (const char *pathname, int algorithm);
 void usage ();
 void help ();
 BOOL is_rebaseable (const char *pathname);
@@ -69,9 +75,11 @@ WORD machine = IMAGE_FILE_MACHINE_I386;
 ULONG64 image_base = 0;
 ULONG64 low_addr;
 BOOL down_flag = FALSE;
+BOOL checksum_flag = FALSE;
 BOOL image_info_flag = FALSE;
 BOOL image_storage_flag = FALSE;
 BOOL image_oblivious_flag = FALSE;
+BOOL perform_rebase_flag = TRUE;
 BOOL force_rebase_flag = FALSE;
 ULONG offset = 0;
 int args_index = 0;
@@ -258,9 +266,15 @@ main (int argc, char *argv[])
       ULONG64 new_image_base = image_base;
       for (i = 0; i < img_info_size; ++i)
 	{
+	  int compactos_algorithm
+	      = compactos_get_algorithm (img_info_list[i].name);
 	  status = rebase (img_info_list[i].name, &new_image_base, down_flag);
 	  if (!status)
 	    return 2;
+	  /* Reapply previous compression. */
+	  if (compactos_algorithm >= 0)
+	    compactos_compress_file (img_info_list[i].name,
+				     compactos_algorithm);
 	}
     }
   else
@@ -268,6 +282,9 @@ main (int argc, char *argv[])
       /* Rebase with database support. */
       BOOL header;
 
+#if defined(__CYGWIN__)
+      int compactos_algorithm = compactos_get_algorithm ("/bin/cygwin1.dll");
+#endif
       if (merge_image_info () < 0)
 	return 2;
       status = TRUE;
@@ -278,6 +295,14 @@ main (int argc, char *argv[])
 	    status = rebase (img_info_list[i].name, &new_image_base, FALSE);
 	    if (status)
 	      img_info_list[i].flag.needs_rebasing = 0;
+#if defined(__CYGWIN__)
+	    /* If Cygwin DLL is compressed, assume setup was used with option
+	       --compact-os.  Align compression with Cygwin DLL. */
+	    if (compactos_algorithm >= 0
+		&& compactos_compress_file (img_info_list[i].name,
+					    compactos_algorithm) < 0)
+	      compactos_algorithm = -1;
+#endif
 	  }
       for (header = FALSE, i = 0; i < img_info_size; ++i)
 	if (img_info_list[i].flag.cannot_rebase == 1)
@@ -301,6 +326,10 @@ main (int argc, char *argv[])
 	      }
 	    fprintf (stderr, "  %s\n", img_info_list[i].name);
 	  }
+      /* Print list of DLLs which still overlap.  This could occur if DLLs are
+	 not rebaseable or if --merge-files is used incorrectly. */
+      if (img_info_size)
+	print_overlapped ();
       if (save_image_info () < 0)
 	return 2;
     }
@@ -533,9 +562,9 @@ load_image_info ()
       {
 	img_info_list[i].name = NULL;
 	/* Ensure that existing database entries are not touched when
-	 *  --oblivious is active, even if they are out-of sync with
-	 *  reality. */
-	if (image_oblivious_flag)
+	 *  --oblivious or --merge-files is active, even if they are
+	 *  out-of sync with reality. */
+	if (image_oblivious_flag || !perform_rebase_flag)
 	  img_info_list[i].flag.cannot_rebase = 2;
       }
   /* Eventually read the strings. */
@@ -584,10 +613,11 @@ load_image_info ()
 static BOOL
 set_cannot_rebase (img_info_t *img)
 {
-  /* While --oblivious is active, cannot_rebase is set to 2 on loading
-   * the database entries */
+  /* While --oblivious or --merge-files is active, cannot_rebase
+   * is set to 2 on loading the database entries */
   if (img->flag.cannot_rebase <= 1 )
     {
+      /* This also removes Compact OS compression. */
       int fd = open (img->name, O_WRONLY);
       if (fd < 0)
 	img->flag.cannot_rebase = 1;
@@ -602,6 +632,7 @@ merge_image_info ()
 {
   int i, end;
   img_info_t *match;
+  BOOL sorted;
   ULONG64 floating_image_base;
 
   /* Sort new files from command line by name. */
@@ -649,7 +680,14 @@ merge_image_info ()
 		{
 		  /* Reuse the old address if possible. */
 		  if (match->slot_size < img_info_list[i].slot_size)
-		    match->base = 0;
+		    {
+		      match->base = 0;
+		      if (verbose)
+		        fprintf (stderr, "rebasing %s because it won't fit in it's old slot size\n", img_info_list[i].name);
+		    }
+		  else if (verbose)
+		    fprintf (stderr, "rebasing %s because it's not located at it's old slot\n", img_info_list[i].name);
+
 		  match->flag.needs_rebasing = 1;
 		}
 	      /* Unconditionally overwrite old with new size. */
@@ -668,8 +706,23 @@ merge_image_info ()
 	      img_info_list[i--] = img_info_list[--img_info_size];
 	    }
 	  else if (!img_info_list[i].flag.cannot_rebase)
-	    /* Not in database yet.  Set base to 0 to choose a new one. */
-	    img_info_list[i].base = 0;
+	    {
+	      /* Not in database yet.  Set base to 0 to choose a new one. */
+	      img_info_list[i].base = 0;
+	      if (verbose)
+		fprintf (stderr, "rebasing %s because not in database yet\n", img_info_list[i].name);
+	    }
+	  else if (img_info_list[i].flag.needs_rebasing)
+	    {
+	      /* Not in database yet and not rebaseable.  Add without rebasing or
+	         skip if --oblivious is active. */
+	      img_info_list[i].flag.needs_rebasing = 0;
+	      if (verbose)
+		fprintf (stderr, "not rebasing %s because file is not writable\n",
+			 img_info_list[i].name);
+	      /* FIXME: Overlaps of DLLs in the database with this DLL will
+	         not be detected below. */
+	    }
 	}
     }
   if (!img_info_rebase_start || force_rebase_flag)
@@ -680,78 +733,146 @@ merge_image_info ()
 	{
 	  /* Test DLLs already in database for writability. */
 	  if (i < img_info_rebase_start)
-	    set_cannot_rebase (&img_info_list[i]);
+	    {
+	      set_cannot_rebase (&img_info_list[i]);
+	      if (img_info_list[i].flag.cannot_rebase == 1 && verbose)
+		fprintf (stderr, "not rebasing %s because file is not writable\n",
+			 img_info_list[i].name);
+	    }
 	  if (!img_info_list[i].flag.cannot_rebase)
-	    img_info_list[i].base = 0;
+	    {
+	      img_info_list[i].base = 0;
+	      if (verbose)
+		fprintf (stderr, "rebasing %s because forced or database missing\n", img_info_list[i].name);
+	    }
 	}
       img_info_rebase_start = 0;
     }
 
-  /* Now sort the old part of the list by base address. */
-  if (img_info_rebase_start)
-    qsort (img_info_list, img_info_rebase_start, sizeof (img_info_t),
-	   img_info_cmp);
-  /* Perform several tests on the information fetched from the database
-     to match with reality. */
-  for (i = 0; i < img_info_rebase_start; ++i)
+  for (sorted = FALSE; img_info_rebase_start && !sorted; )
     {
-      ULONG64 cur_base;
-      ULONG cur_size, slot_size;
+      char overlaps[img_info_rebase_start];
+      memset(overlaps, 0, img_info_rebase_start);
+      /* Now sort the old part of the list by base address. */
+      qsort (img_info_list, img_info_rebase_start, sizeof (img_info_t),
+	     img_info_cmp);
+      /* Perform several tests on the information fetched from the database
+	 to match with reality. */
+      for (sorted = TRUE, i = 0; sorted && i < img_info_rebase_start; ++i)
+	{
+	  ULONG64 cur_base, cur_base_orig;
+	  ULONG cur_size, slot_size;
 
-      /* Files with the needs_rebasing or cannot_rebase flags set have been
-	 checked already. */
-      if (img_info_list[i].flag.needs_rebasing
-      	  || img_info_list[i].flag.cannot_rebase)
-	continue;
-      /* Check if the files in the old list still exist.  Drop non-existant
-	 or unaccessible files. */
-      if (access (img_info_list[i].name, F_OK) == -1
-	  || !GetImageInfos64 (img_info_list[i].name, NULL,
-			       &cur_base, &cur_size))
-	{
-	  free (img_info_list[i].name);
-	  memmove (img_info_list + i, img_info_list + i + 1,
-		   (img_info_size - i - 1) * sizeof (img_info_t));
-	  --img_info_rebase_start;
-	  --img_info_size;
-	  continue;
-	}
-      slot_size = roundup2 (cur_size, ALLOCATION_SLOT);
-      if (set_cannot_rebase (&img_info_list[i]))
-	img_info_list[i].base = cur_base;
-      else
-	{
+	  /* If file is not rebaseable and neither --oblivious nor --merge-files
+	     is active, search forward for possible overlaps. */
+	  if (img_info_list[i].flag.cannot_rebase == 1)
+	    {
+	      int j;
+	      /* If cannot_rebase is set, base and size are already valid. */
+	      cur_base = img_info_list[i].base;
+	      slot_size = img_info_list[i].slot_size;
+	      for (j = i + 1; j < img_info_rebase_start; ++j)
+		{
+		  if (img_info_list[j].flag.cannot_rebase)
+		    continue; /* Also not rebaseable. */
+		  if (img_info_list[j].base == 0)
+		    continue; /* Will already be rebased from scratch. */
+		  if (cur_base + slot_size + offset <= img_info_list[j].base)
+		    break; /* This and any further DLLs do not overlap. */
+		  /* Select rebase from scratch when outer loop arrives at j. */
+		  overlaps[j] = 1;
+		}
+	    }
+
+	  /* Files with the needs_rebasing or cannot_rebase flags set have been
+	     checked already. */
+	  if (img_info_list[i].flag.needs_rebasing
+	      || img_info_list[i].flag.cannot_rebase)
+	    continue;
+	  /* Check if the files in the old list still exist.  Drop non-existant
+	     or unaccessible files. */
+	  if (access (img_info_list[i].name, F_OK) == -1
+	      || !GetImageInfos64 (img_info_list[i].name, NULL,
+				   &cur_base, &cur_size))
+	    {
+	      free (img_info_list[i].name);
+	      memmove (img_info_list + i, img_info_list + i + 1,
+		       (img_info_size - i - 1) * sizeof (img_info_t));
+	      memmove (overlaps + i, overlaps + i + 1, img_info_size - i - 1);
+	      --img_info_rebase_start;
+	      --img_info_size;
+	      --i;
+	      continue;
+	    }
+	  slot_size = roundup2 (cur_size, ALLOCATION_SLOT);
+	  cur_base_orig = cur_base;
 	  /* If the file has been reinstalled, try to rebase to the same address
 	     in the first place. */
 	  if (cur_base != img_info_list[i].base)
 	    {
 	      img_info_list[i].flag.needs_rebasing = 1;
+	      if (verbose)
+		fprintf (stderr, "rebasing %s because it's base has changed "
+			 "(due to being reinstalled?)\n", img_info_list[i].name);
 	      /* Set cur_base to the old base to simplify subsequent tests. */
 	      cur_base = img_info_list[i].base;
 	    }
 	  /* However, if the DLL got bigger and doesn't fit into its slot
 	     anymore, rebase this DLL from scratch. */
 	  if (i + 1 < img_info_rebase_start
-	      && cur_base + slot_size + offset >= img_info_list[i + 1].base)
-	    img_info_list[i].base = 0;
-	  /* Does the previous DLL reach into the address space of this
+	      && cur_base + slot_size + offset > img_info_list[i + 1].base)
+	    {
+	      img_info_list[i].base = 0;
+	      if (verbose)
+		fprintf (stderr, "rebasing %s because it won't fit in it's old slot "
+			 "without overlapping next DLL\n", img_info_list[i].name);
+	    }
+	  /* Does a previous DLL reach into the address space of this
 	     DLL?  This happens if the previous DLL is not rebaseable. */
-	  else if (i > 0 && cur_base < img_info_list[i - 1].base
-				       + img_info_list[i + 1].slot_size)
-	    img_info_list[i].base = 0;
+	  else if (overlaps[i])
+	    {
+	      img_info_list[i].base = 0;
+	      if (verbose)
+		fprintf (stderr, "rebasing %s because a previous non-writable DLL "
+			 "overlaps\n", img_info_list[i].name);
+	    }
 	  /* Does the file match the base address requirements?  If not,
 	     rebase from scratch. */
-	  else if ((down_flag && cur_base + slot_size + offset >= image_base)
+	  else if ((down_flag && cur_base + slot_size + offset > image_base)
 		   || (!down_flag && cur_base < image_base))
-	    img_info_list[i].base = 0;
+	    {
+	      img_info_list[i].base = 0;
+	      if (verbose)
+		fprintf (stderr, "rebasing %s because it's base address is outside the "
+			 "expected area\n", img_info_list[i].name);
+	    }
+	  /* Make sure all DLLs with base address 0 have the needs_rebasing
+	     flag set. */
+	  if (img_info_list[i].base == 0)
+	    img_info_list[i].flag.needs_rebasing = 1;
+	  /* Only check for writability if file needs rebasing to keep
+	     Compact OS compression of unchanged files.  Revert rebase
+	     decision if not writeable. */
+	  if (img_info_list[i].flag.needs_rebasing
+	      && set_cannot_rebase (&img_info_list[i]))
+	    {
+	      img_info_list[i].flag.needs_rebasing = 0;
+	      img_info_list[i].base = cur_base_orig;
+	      if (verbose)
+		fprintf (stderr, "rebasing %s deferred because file is not writable\n",
+			 img_info_list[i].name);
+	      /* List is possibly no longer sorted which could result in false
+	         negative rebase decisions.  Exit loop and restart it with
+		 newly sorted list.  This will also set the overlaps[] array
+		 for this DLL when the restarted loop visits the (possibly new)
+		 index of this DLL.  Infinite retries could not occur because
+		 each turn sets another cannot_rebase flag. */
+	      sorted = FALSE;
+	    }
+	  /* Unconditionally overwrite old with new size. */
+	  img_info_list[i].size = cur_size;
+	  img_info_list[i].slot_size = slot_size;
 	}
-      /* Unconditionally overwrite old with new size. */
-      img_info_list[i].size = cur_size;
-      img_info_list[i].slot_size = slot_size;
-      /* Make sure all DLLs with base address 0 have the needs_rebasing
-	 flag set. */
-      if (img_info_list[i].base == 0)
-	img_info_list[i].flag.needs_rebasing = 1;
     }
   /* The remainder of the function expects img_info_size to be > 0. */
   if (img_info_size == 0)
@@ -849,7 +970,7 @@ collect_image_info (const char *pathname)
 
   /* Skip if not rebaseable, but only if we're collecting for rebasing,
      not if we're collecting for printing only. */
-  if (!image_info_flag && !is_rebaseable (pathname))
+  if (perform_rebase_flag && !is_rebaseable (pathname))
     {
       if (!quiet)
 	fprintf (stderr, "%s: skipped because not rebaseable\n", pathname);
@@ -899,8 +1020,16 @@ collect_image_info (const char *pathname)
     }
   img_info_list[img_info_size].slot_size
     = roundup2 (img_info_list[img_info_size].size, ALLOCATION_SLOT);
-  img_info_list[img_info_size].flag.needs_rebasing = 1;
-  img_info_list[img_info_size].flag.cannot_rebase = 0;
+  if (perform_rebase_flag)
+    {
+      img_info_list[img_info_size].flag.needs_rebasing = 1;
+      img_info_list[img_info_size].flag.cannot_rebase = 0;
+    }
+  else
+    {
+      img_info_list[img_info_size].flag.needs_rebasing = 0;
+      img_info_list[img_info_size].flag.cannot_rebase = 2;
+    }
   /* This back and forth from POSIX to Win32 is a way to get a full path
      more thoroughly.  For instance, the difference between /bin and
      /usr/bin will be eliminated. */
@@ -940,6 +1069,10 @@ collect_image_info (const char *pathname)
     img_info_list[img_info_size].name_size = strlen (full_path) + 1;
   }
 #endif
+  if (verbose)
+    fprintf (stderr, "%s %s because filename given on command line\n",
+	     perform_rebase_flag ? "rebasing" : "considering",
+	     img_info_list[img_info_size].name);
   ++img_info_size;
   return TRUE;
 }
@@ -1024,12 +1157,72 @@ print_image_info ()
     }
 }
 
+static void
+print_overlapped ()
+{
+  BOOL header;
+  int i;
+  char overlaps[img_info_size];
+  memset (overlaps, 0, img_info_size);
+  qsort (img_info_list, img_info_size, sizeof (img_info_t), img_info_cmp);
+  for (header = FALSE, i = 0; i < img_info_size; ++i)
+    {
+      int j;
+      if (img_info_list[i].flag.needs_rebasing)
+	continue; /* Rebase failed. */
+      for (j = i + 1; j < img_info_size; ++j)
+	{
+	  if (img_info_list[j].flag.needs_rebasing)
+	    continue; /* Rebase failed. */
+	  if (img_info_list[i].base + img_info_list[i].slot_size + offset
+	      <= img_info_list[j].base)
+	    break;
+	  overlaps[i] = overlaps[j] = 1;
+	}
+      if (!overlaps[i])
+	continue;
+      if (!header)
+	{
+	  fputs ("\nThe following DLLs still overlap:\n", stderr);
+	  header = TRUE;
+	}
+      fprintf (stderr, "  %s\n", img_info_list[i].name);
+    }
+}
+
+static int
+update_checksum (const char *pathname)
+{
+  int fd, err;
+  unsigned old_checksum, new_checksum;
+
+  if ((fd = open (pathname, O_RDWR | O_BINARY)) == -1)
+    {
+      fprintf (stderr, "%s: failed to reopen for checksum update: %s\n",
+	       pathname, strerror (errno));
+      return -1;
+    }
+  new_checksum = pe32_checksum (fd, 1, &old_checksum);
+  err = errno;
+  close(fd);
+  if (!new_checksum)
+    {
+      fprintf (stderr, "%s: checksum update failed: %s\n", pathname,
+	       strerror (err));
+      /* Assume file is unchanged. */
+      return -1;
+    }
+
+  return (new_checksum != old_checksum ? 1 : 0);
+}
+
 BOOL
 rebase (const char *pathname, ULONG64 *new_image_base, BOOL down_flag)
 {
   ULONG64 old_image_base, prev_new_image_base;
   ULONG old_image_size, new_image_size;
   BOOL status;
+  const char *checksum_msg;
 
   /* Skip if not writable. */
   if (access (pathname, W_OK) == -1)
@@ -1039,6 +1232,9 @@ rebase (const char *pathname, ULONG64 *new_image_base, BOOL down_flag)
       return TRUE;
     }
 
+  /* Try to update the checksum. */
+  ReBaseUpdateCheckSum = TRUE;
+
   /* Calculate next base address, if rebasing down. */
   if (down_flag)
     *new_image_base -= offset;
@@ -1115,13 +1311,25 @@ retry:
     }
 #endif
 
+
+  /* Fix checksum, if requested or incremental update failed. */
+  if (checksum_flag || !ReBaseUpdateCheckSum)
+    switch (update_checksum (pathname))
+      {
+	case 0:  checksum_msg = ", checksum verified"; break;
+	case 1:  checksum_msg = ", checksum fixed"; break;
+	default: checksum_msg = ", checksum update failed"; break;
+      }
+  else
+    checksum_msg = "";
+
   /* Display rebase results, if verbose. */
   if (verbose)
     {
-      printf ("%s: new base = %" PRIx64 ", new size = %x\n",
+      printf ("%s: new base = %" PRIx64 ", new size = %x%s\n",
 	      pathname,
 	      (uint64_t) ((down_flag) ? *new_image_base : prev_new_image_base),
-	      (uint32_t) (new_image_size + offset));
+	      (uint32_t) (new_image_size + offset), checksum_msg);
     }
 
   /* Calculate next base address, if rebasing up. */
@@ -1135,10 +1343,12 @@ static struct option long_options[] = {
   {"32",	no_argument,	   NULL, '4'},
   {"64",	no_argument,	   NULL, '8'},
   {"base",	required_argument, NULL, 'b'},
+  {"checksum",	no_argument,	   NULL, 'c'},
   {"down",	no_argument,	   NULL, 'd'},
   {"help",	no_argument,	   NULL, 'h'},
   {"usage",	no_argument,	   NULL, 'h'},
   {"info",	no_argument,	   NULL, 'i'},
+  {"merge-files",no_argument,	   NULL, 'M'},
   {"offset",	required_argument, NULL, 'o'},
   {"oblivious",	no_argument,	   NULL, 'O'},
   {"quiet",	no_argument,	   NULL, 'q'},
@@ -1151,7 +1361,7 @@ static struct option long_options[] = {
   {NULL,	no_argument,	   NULL,  0 }
 };
 
-static const char *short_options = "48b:dhino:OqstT:vV";
+static const char *short_options = "48b:cdhiMno:OqstT:vV";
 
 void
 parse_args (int argc, char *argv[])
@@ -1176,11 +1386,15 @@ parse_args (int argc, char *argv[])
 	  image_base = string_to_ulonglong (optarg);
 	  force_rebase_flag = TRUE;
 	  break;
+	case 'c':
+	  checksum_flag = TRUE;
+	  break;
 	case 'd':
 	  down_flag = TRUE;
 	  break;
 	case 'i':
 	  image_info_flag = TRUE;
+	  perform_rebase_flag = FALSE;
 	  break;
 	case 'o':
 	  offset = string_to_ulonglong (optarg);
@@ -1189,6 +1403,10 @@ parse_args (int argc, char *argv[])
 	case 'q':
 	  quiet = TRUE;
 	  break;
+	case 'M':
+	  perform_rebase_flag = FALSE;
+	  image_storage_flag = TRUE;
+	  break;
 	case 'O':
 	  image_oblivious_flag = TRUE;
 	  /* -O implies -s, which in turn implies -d, so intentionally
@@ -1233,8 +1451,8 @@ parse_args (int argc, char *argv[])
 	}
     }
 
-  if ((image_base == 0 && !image_info_flag && !image_storage_flag)
-      || (image_base && image_info_flag))
+  if ((image_base == 0 && perform_rebase_flag && !image_storage_flag)
+      || (force_rebase_flag && !perform_rebase_flag))
     {
       usage ();
       exit (1);
@@ -1332,13 +1550,86 @@ string_to_ulonglong (const char *string)
   return number;
 }
 
+static int
+compactos_get_algorithm (const char *pathname)
+{
+  /* Requires Win10. */
+  if (!IsWindows10OrGreater ())
+    return -1;
+
+  int fd = open (pathname, O_RDONLY);
+  if (fd == -1)
+    return -1;
+
+  struct {
+    WOF_EXTERNAL_INFO Wof;
+    FILE_PROVIDER_EXTERNAL_INFO_V1 FileProvider;
+  } wfp = { {0, 0}, {0, 0, 0} };
+
+  int rc;
+  if (!DeviceIoControl ((HANDLE) _get_osfhandle (fd),
+			FSCTL_GET_EXTERNAL_BACKING, NULL, 0,
+			&wfp, sizeof(wfp), NULL, NULL))
+    rc = -1;
+  else
+    rc = (wfp.Wof.Provider == WOF_PROVIDER_FILE ?
+	  wfp.FileProvider.Algorithm : -1);
+
+  close (fd);
+  return rc;
+}
+
+static int
+compactos_compress_file (const char *pathname, int algorithm)
+{
+  /* Do not apply unknown algorithms. */
+  if (!(FILE_PROVIDER_COMPRESSION_XPRESS4K <= algorithm &&
+        algorithm <= FILE_PROVIDER_COMPRESSION_XPRESS16K))
+    return 0;
+
+  int fd = open (pathname, O_RDONLY);
+  if (fd == -1)
+    return -1;
+  HANDLE h = (HANDLE) _get_osfhandle (fd);
+
+  /* Older versions of Win10 set mtime to current time. */
+  FILETIME ft;
+  BOOL ft_valid = GetFileTime (h, NULL, NULL, &ft);
+
+  struct {
+    WOF_EXTERNAL_INFO Wof;
+    FILE_PROVIDER_EXTERNAL_INFO_V1 FileProvider;
+  } wfp;
+  wfp.Wof.Version = WOF_CURRENT_VERSION;
+  wfp.Wof.Provider = WOF_PROVIDER_FILE;
+  wfp.FileProvider.Version = FILE_PROVIDER_CURRENT_VERSION;
+  wfp.FileProvider.Algorithm = algorithm;
+  wfp.FileProvider.Flags = 0;
+
+  int rc;
+  if (!DeviceIoControl (h, FSCTL_SET_EXTERNAL_BACKING, &wfp, sizeof(wfp),
+			NULL, 0, NULL, NULL))
+    rc = (GetLastError() == ERROR_COMPRESSION_NOT_BENEFICIAL ? 0 : -1);
+  else
+    rc = 1;
+
+  if (ft_valid)
+    SetFileTime (h, NULL, NULL, &ft);
+
+  close (fd);
+  if (verbose)
+    printf ("%s: Compact OS algorithm %d %s\n", pathname, algorithm,
+	    (rc < 0 ? "FAILED" : rc == 0 ? "not applied " : "applied"));
+  return rc;
+}
+
 void
 usage ()
 {
   fprintf (stderr,
 "usage: %s [-b BaseAddress] [-o Offset] [-48dOsvV]"
 " [-T [FileList | -]] Files...\n"
-"       %s -i [-48Os] [-T [FileList | -]] Files...\n"
+"       %s -i [-48MOs] [-T [FileList | -]] Files...\n"
 "       %s --help or --usage for full help text\n",
 	   progname, progname, progname);
 }
@@ -1368,15 +1659,26 @@ Rebase PE files, usually DLLs, to a specified address or address range.\n\
   -O, --oblivious         Do not change any files already in the database\n\
                           and do not record any changes to the database.\n\
                           (Implies -s).\n\
+  -M, --merge-files       Do not perform any rebase, update the database only.\n\
+                          Implies -s, incompatible with -b and -o.\n\
+                          The files listed should have been rebased just before\n\
+                          using the -O flag (maybe in some staging directory).\n\
   -i, --info              Rather then rebasing, just print the current base\n\
                           address and size of the files.  With -s, use the\n\
                           database.  The files are ordered by base address.\n\
                           A '*' at the end of the line is printed if a\n\
                           collisions with an adjacent file is detected.\n\
+                          Incompatible with -b and -o.\n\
 \n\
   One of the options -b, -s or -i is mandatory.  If no rebase database exists\n\
   yet, -b is required together with -s.\n\
 \n\
+  -c, --checksum          Fix the file's checksum in the PE header if the file\n\
+                          has been successfully rebased.  This also bumps the\n\
+                          file's modification time (like -t) if the checksum\n\
+                          has been changed.  By default, the checksum is\n\
+                          updated without reading the whole file under the\n\
+                          assumption that the original checksum was correct.\n\
   -d, --down              Treat the BaseAddress as upper ceiling and rebase\n\
                           files top-down from there.  Without this option the\n\
                           files are rebased from BaseAddress bottom-up.\n\
@@ -1385,8 +1687,10 @@ Rebase PE files, usually DLLs, to a specified address or address range.\n\
   -o, --offset=OFFSET     Specify an additional offset between adjacent DLLs\n\
                           when rebasing.  Default is no offset.\n\
   -t, --touch             Use this option to make sure the file's modification\n\
-                          time is bumped if it has been successfully rebased.\n\
-                          Usually rebase does not change the file's time.\n\
+                          time and the timestamp in the PE header are bumped if\n\
+                          the file has been successfully rebased.\n\
+                          Usually rebase does not change the file's time unless\n\
+                          the -c flag is also specified.\n\
   -T, --filelist=FILE     Also rebase the files specified in FILE.  The format\n\
                           of FILE is one DLL per line.\n\
   -q, --quiet             Be quiet about non-critical issues.\n\
@@ -1410,7 +1714,7 @@ is_rebaseable (const char *pathname)
 
   fd = open (pathname, O_RDONLY);
   if (fd == -1)
-    goto done;
+    return status;
 
   offset = lseek (fd, pe_signature_offset_offset, SEEK_SET);
   if (offset == -1)
